SOAPAction =
  #
  # Constructor for a SOAPAction object typically
  #  read from a WSDL file.
function(x, obj = new("SOAPAction"))
{
  as(obj, "character") = x

  obj
}


SOAPType =
  #
  # 
  #
function(name, ns = "", nsuri="", obj = new("SOAPType"))
{

  if(is.null(name)) {
    return(new("SOAPVoidType"))
  }

   # If the name is in the form of namespace:type, split it
   # and assign to name and ns separately.
   # e.g. xsd:int will be come name = "int", ns = "xsd"
  name = strsplit(name, ":")[[1]]
  if(length(name) == 2) {
    ns = name[1]
    name = name[2]
  }

  if(ns == "xsd") {
    which = sapply(SOAPTypes, function(x) paste(ns, name, sep=":") == x)
    if(any(which)) {
       obj = new("PrimitiveSOAPType")
       obj@toConverter = get(paste("as", names(SOAPTypes)[which], sep="."))
    }
  } else {
     # Should check that this is referring to the target namespace of the
     # document, e.g. targetNamespace and xmlns:[targetNamespace] in the root node
     obj = new("SOAPTypeReference")
  }


#  if(is(obj, "SOAPType"))
#    browser()
  
  obj@name = name
  obj@ns = ns
  obj@nsuri = nsuri

  obj
}

SOAPServerDescription =
function(name,  server, operations, types, obj = new("SOAPServerDescription"))
{
 obj@name = name
 obj@server = server
 if(!is.null(operations))
   obj@operations = operations

 if(!is.null(types))
   obj@types = types

 obj
}


processWSDL <-
  #
  # Currently only handle one <service> element
  #
  function(fileName = "KEGG.wsdl")
{
 library(XML)
  # Parse the WSDL file into a tree which we will pick apart.
  # Could put handlers to convert the elements on the fly.
 doc = xmlRoot(xmlTreeParse(fileName, handlers = list(comment = function(x) NULL), asTree = TRUE))

  # Get the (first) server declared in <service><port>...
 port = doc[["service"]][["port"]]

 loc = xmlGetAttr(port[["address"]], "location")

 server = SOAPServer(loc)
 
  # Now find the methods/functions that are available from this server.
  # These are in the <binding> elements. We should match each one to
  # each <service>. Currently only handle one <service> element
 
 tmp = doc[names(doc) == "binding"]
 ops = lapply(tmp, processWSDLBindings, doc)
 names(ops) = sapply(tmp, function(x) xmlGetAttr(x, "name"))

 types = processWSDLTypes(doc[["types"]], doc)


 SOAPServerDescription(name = xmlGetAttr(port, "name"), server = server, operations = ops, types = types)
}

processWSDLTypes =
   # Process all the elements in types (hopefully <schema> elements)
   # and merge. We may want to keep these as a list indexed by namespace/schema identifier.
 # Originally, we just processed the first one.
 # types = processWSDLTypes(doc[["types"]][[1]], doc)   
function(node, doc)
{
  if(is.null(node))
      return(list())

   # Find the URIs of the namespaces for the different schema
 ns = xmlSApply(node, function(x) xmlGetAttr(x, "targetNamespace"))
   # map these to namespace identifiers/short-names
 nsDefs = doc$namespaceDefinitions
 which = match(ns, sapply(nsDefs, function(x) x$uri))
 ns = names(nsDefs)[which]
 types = lapply(1:xmlSize(doc[["types"]]),
                function(i) {
                   processSchemaTypes(doc[["types"]][[i]], doc, ns = ns[i])
                 })
  # Fix the names on these types to avoid the schema.
 types = unlist(types, recursive = FALSE)
 names(types) = gsub("^schema\.", "", names(types))

 types
}


processSchemaTypes =
function(types,  doc, ns = character(0))
{

 if(is.null(types))
   return(NULL)

  # XXX need to handle simple types and simple content.
 tmp = xmlApply(types, function(x, ...) if(xmlName(x) %in% c("complexType", "complexContent")) processWSDLType(x, ...) else NULL, types)
 names(tmp) = sapply(tmp, function(x) paste(ns, x$name, sep = ":"))

  # Drop the NULL elements which correspond to non-complexType elements,
  # e.g. <import namespace="..">
 tmp = tmp[!sapply(tmp, is.null)]

 tmp  
}


ArrayType =
function(elementType, obj = new("ArrayType"))
{
  obj@elementType = elementType

  obj
}  

ClassDef =
function(name, slotTypes, obj = new("ClassDefinition"))
{
  obj@name = name
  obj@slotTypes = slotTypes

  obj
}


# This material relates to XML schema and processing of types defined within
# schema. Ideally, we will separate this into a separate package and use the
# results in different ways, not just for SOAP. And we will provide a more
# extensive framework for dealing with all of the schema details rather than
# the rather limited but common ones supported here.
#
#  See the O'Reilly book XML Schemas by Eric van der Vlist for more information
#  on schemas, or look on the Web (e.g. www.xml.com/schemas)
#
processWSDLType =

  #
  # This currently deals with a very small subset of XML schema
  # specifications.  These are the common ones, but more work 
  # is definitely needed.
  #
function(type, types)
{
# Rationalize this 

  name = xmlGetAttr(type, "name")
  
  tmp <- type[["complexContent"]]

  if(!is.null(tmp) && !is.null(tmp[["all"]])) {
      # a struct-like definition with slots.
     tmp = tmp[["all"]]
     slotTypes = xmlApply(tmp, function(x) xmlGetAttr(x, "type"))
     names(slotTypes) = xmlSApply(tmp, function(x) xmlGetAttr(x, "name"))
     def = ClassDef(name, slotTypes)
  } else if(!is.null(tmp) && !is.null(tmp[["extension"]])) {
     base = xmlGetAttr(tmp[["extension"]], "base")
     # Now get the extensions.
     def = ClassDef(name, list(), new("ExtendedClassDefinition"))
     def@base = base
  } else if(!is.null(tmp)) {
    # Currently we interpret this as an Array.
    # So get the <restriction><attribute> element
    # and take the arrayType and ref attribution from that.
    restriction = tmp[["restriction"]]
    a = restriction[["attribute"]]
    def = ArrayType(xmlGetAttr(a, "arrayType"))
  } else if(!is.null(tmp <- type[["all"]])) {
      # a struct-like definition with slots.
     slotTypes = xmlApply(tmp, function(x) xmlGetAttr(x, "type"))
     names(slotTypes) = xmlSApply(tmp, function(x) xmlGetAttr(x, "name"))
     def = ClassDef(name, slotTypes)
  }
  
  list(name = name, definition = def)
}


processWSDLBindings =
  #
  # Called with a <binding>...</binding> node.
  #
function(node, doc = NULL)
{
 els =  node[names(node) == "operation"]
 
 type = strsplit(xmlGetAttr(node, "type"), ":")[[1]][2]
 
 types = doc[xmlSApply(doc, function(x) xmlName(x) == "portType" && xmlGetAttr(x, "name") == type)][[1]]
 
 ops = lapply(els, processWSDLOperation, types, doc)
 names(ops) = sapply(els, function(x) xmlGetAttr(x, "name"))
   
 ops  
}


processWSDLOperation =
  #
  # An operation node.
  #
function(node, types, doc)
{
  name = xmlGetAttr(node, "name")

  which = xmlSApply(types, function(x) xmlGetAttr(x, "name") == name)

  portType = types[which][[1]]
  inputNode = xmlGetAttr(portType[["input"]], "message")
  input = strsplit(inputNode,":")[[1]][2]
  
  msgs = doc[xmlSApply(doc, function(x) xmlName(x) == "message" && xmlGetAttr(x, "name") == input)]
  msg=msgs[[1]]
  args = xmlApply(msg, function(x) SOAPType(xmlGetAttr(x, "type")))
  names(args) = xmlSApply(msg, function(x) xmlGetAttr(x, "name"))

  output = xmlGetAttr(portType[["output"]], "message")
  output = strsplit(output,":")[[1]][2]
  msgs = doc[xmlSApply(doc, function(x) xmlName(x) == "message" && xmlGetAttr(x, "name") == output)]
  msg=msgs[[1]]  

  if(xmlSize(msg))
    value = xmlGetAttr(msg[["part"]], "type")
  else
    value = NULL

  new("WSDLMethod", name = name, parameters = args, returnValue = SOAPType(value),
                    namespace = xmlGetAttr(node[["input"]][["body"]], "namespace", ""),
                    action = SOAPAction(xmlGetAttr(node[["operation"]], "soapAction")))
}  


# Resolve finds the definition and its contents within a context
# It can resolve at one level or all the way down.
# The definitions here are typically SOAP types.

setGeneric("resolve", function(obj, context, recursive = TRUE) standardGeneric("resolve"))

setMethod("resolve", c("BasicSOAPType", "list"),
           function(obj, context, recursive = TRUE) {
             obj
           })

setMethod("resolve", c("SOAPVoidType"),
           function(obj, context, recursive = TRUE) {
             obj
           })

setMethod("resolve", c("SOAPTypeReference", "list"),
           function(obj, context, recursive = TRUE) {
             resolve(obj@name, context, recursive)
           })

setMethod("resolve", c("ArrayType", "list"),
           function(obj, context, recursive = TRUE) {
             if(recursive && (is.null(obj@elType) || length(obj@elType@name) == 0)) {
               name = obj@elementType
               obj@elType = resolve(name, context, recursive)
             }
             obj
           })

setMethod("resolve", c("character", "list"),
           function(obj, context, recursive = TRUE) {

               els = strsplit(gsub("\\\[\\\]", "", obj), ":")[[1]]
               if(length(els) == 2) {
                 if(els[1] == "xsd") {
                   return(SOAPType(els[2], els[1]))
                 }
                 els = els[2]
               } 

             
               el = match(els[1], names(context))
               if(is.na(el)) {
                  el = match(els[1], gsub(".*:", "", names(context)))
               }
               if(is.na(el)) 
                 stop("Cannot resolve SOAPTypeReference in list")

               ans = context[[el]]$definition

               if(recursive && !is(ans, "TerminalSOAPType"))
                 ans = resolve(ans, context, recursive)

               ans
           })

